# pylint: skip-file
import datetime
import logging
import os
import random
import subprocess
import threading
import time

import pytest
from pytest_shell import fs

from opentaskpy import taskrun
from tests.fixtures.ssh_clients import *  # noqa: F403

# Create a variable with a random number
RANDOM = random.randint(10000, 99999)
FILE_PREFIX = "unittest_task_run"
MOVED_FILES_DIR = "archive"
DELIMITER = ","

"""
#################
Tests for the "binary" task runner
#################
"""

# Setup logger so we can see the stdout and err from the binary
logging.basicConfig(
    format="%(asctime)s - %(name)s - %(levelname)s - %(message)s", level=logging.DEBUG
)


def test_noop_binary(env_vars, setup_ssh_keys, root_dir):
    # Pass noop argument to the binary

    # Create a test file
    fs.create_files(
        [{f"{root_dir}/testFiles/ssh_1/src/noop_test.txt": {"content": "test1234"}}]
    )

    assert run_task_run("scp-basic", noop=True)["returncode"] == 0

    # Verify that the file has not been transferred
    assert not os.path.exists(f"{root_dir}/testFiles/ssh_2/dest/noop_test.txt")

    # Override the variables config location and verify we get an error
    assert run_task_run("scp-basic", config="/tmp/non-existent")["returncode"] == 1

    # Verify a --noop runs for an execution too

    # Use the touch example
    touched_file = f"{root_dir}/testFiles/ssh_1/src/touchedFile.txt"
    # Delete if it already exists
    if os.path.exists(touched_file):
        os.remove(touched_file)

    assert run_task_run("touch", noop=True)["returncode"] == 0
    # Verify the file still doesn't exist
    assert not os.path.exists(touched_file)


def test_scp_basic_binary(env_vars, setup_ssh_keys, root_dir):
    # Use the "binary" to trigger the job with command line arguments

    # Required files for this test:
    # ssh_1 : test/testFiles/ssh_1/src/.*\.txt

    # Create a test file
    fs.create_files(
        [{f"{root_dir}/testFiles/ssh_1/src/text.txt": {"content": "test1234"}}]
    )

    assert run_task_run("scp-basic")["returncode"] == 0


def test_scp_basic_no_error_on_exit_binary(env_vars, setup_ssh_keys, root_dir):
    # Use the "binary" to trigger the job with command line arguments

    assert run_task_run("scp-basic-no-error")["returncode"] == 0


def test_scp_basic_no_error_on_exit_binary_1(env_vars, setup_ssh_keys, root_dir):
    # Use the "binary" to trigger the job with command line arguments

    assert run_task_run("scp-basic-no-error-1")["returncode"] == 0


def test_execution_basic_binary(env_vars, setup_ssh_keys, root_dir):
    # Use the "binary" to trigger the job with command line arguments

    assert run_task_run("df")["returncode"] == 0


def test_execution_invalid_host(env_vars, setup_ssh_keys, root_dir):
    # Use the "binary" to trigger the job with command line arguments

    assert run_task_run("df-invalid-host")["returncode"] == 1


def test_batch_basic_binary(env_vars, setup_ssh_keys, root_dir):
    # Use the "binary" to trigger the job with command line arguments

    # Create a test file
    fs.create_files(
        [{f"{root_dir}/testFiles/ssh_1/src/test.txt": {"content": "test1234"}}]
    )

    assert run_task_run("batch-basic")["returncode"] == 0


def test_transfer_local_binary(env_vars, root_dir):
    # Create a test_basic_local.txt file in /tmp/src
    fs.create_files([{"/tmp/src/test_basic_local.txt": {"content": "test1234"}}])
    # Ensure /tmp/dest exists
    if not os.path.exists("/tmp/dest"):
        os.makedirs("/tmp/dest")

    # Use the "binary" to trigger the job with command line arguments
    assert run_task_run("local-basic")["returncode"] == 0


def test_batch_execution_invalid_host(env_vars, setup_ssh_keys, root_dir):
    # Use the "binary" to trigger the job with command line arguments

    assert run_task_run("batch-basic-invalid-execution-host")["returncode"] == 1


def test_binary_invalid_config_file(env_vars, setup_ssh_keys, root_dir):
    # Use the "binary" to trigger the job with command line arguments

    assert run_task_run("scp-basic-non-existent")["returncode"] == 1
    # Check the output indicates that the task could not be found
    assert "Couldn't find task with name: scp-basic-non-existent" in (
        run_task_run("scp-basic-non-existent")["stderr"]
    )


def test_binary_invalid_config_directory(env_vars, setup_ssh_keys, root_dir):
    # Use the "binary" to trigger the job with command line arguments

    result = run_task_run("scp-basic", config="/tmp/non-existent")
    assert result["returncode"] == 1
    # Check the output indicates that no variables could be loaded
    assert "Couldn't find any variables" in (result["stderr"])


"""
#################
Tests using the Python code directly
#################
"""


def test_unknown_task_name(env_vars, setup_ssh_keys, root_dir):
    task_runner = taskrun.TaskRun("non-existent", "test/cfg")

    # Verify an exception with appropriate text is thrown
    with pytest.raises(FileNotFoundError) as e:
        task_runner.run()
    assert e.value.args[0] == "Couldn't find task with name: non-existent"


def test_batch_basic(env_vars, setup_ssh_keys, root_dir):
    # Create a test file
    fs.create_files(
        [{f"{root_dir}/testFiles/ssh_1/src/test.txt": {"content": "test1234"}}]
    )

    # Use the TaskRun class to trigger the job properly
    task_runner = taskrun.TaskRun("batch-basic", "test/cfg")
    assert task_runner.run()


def test_execution_basic(env_vars, setup_ssh_keys, root_dir):
    # Use the TaskRun class to trigger the job properly
    task_runner = taskrun.TaskRun("df", "test/cfg")
    assert task_runner.run()


def test_execution_fail(env_vars, setup_ssh_keys, root_dir):
    # Use the TaskRun class to trigger the job properly
    task_runner = taskrun.TaskRun("fail-command", "test/cfg")
    assert not task_runner.run()


def test_scp_basic(env_vars, setup_ssh_keys, root_dir):
    # Required files for this test:
    # ssh_1 : test/testFiles/ssh_1/src/.*\.txt

    # Create a test file
    fs.create_files(
        [{f"{root_dir}/testFiles/ssh_1/src/test.txt": {"content": "test1234"}}]
    )

    # Use the TaskRun class to trigger the job properly
    task_runner = taskrun.TaskRun("scp-basic", "test/cfg")
    assert task_runner.run()


def test_scp_basic_multiple_dests(env_vars, setup_ssh_keys, root_dir):
    # Required files for this test:
    # ssh_1 : test/testFiles/ssh_1/src/.*\.txt

    # Create a test file
    fs.create_files(
        [{f"{root_dir}/testFiles/ssh_1/src/test.txt": {"content": "test1234"}}]
    )

    # Use the TaskRun class to trigger the job properly
    task_runner = taskrun.TaskRun("scp-basic-multiple-dests", "test/cfg")
    assert task_runner.run()

    # Check the files were copied to all 3 destinations
    assert os.path.exists(f"{root_dir}/testFiles/ssh_2/dest/test.txt")
    assert os.path.exists(f"{root_dir}/testFiles/ssh_2/dest/test-2.txt")
    assert os.path.exists(f"{root_dir}/testFiles/ssh_2/dest/test-3.txt")


def test_scp_basic_10_files(env_vars, setup_ssh_keys, root_dir):
    # Required files for this test:
    # ssh_1 : test/testFiles/ssh_1/src/.*\.txt

    # Create 10 test files
    for i in range(10):
        fs.create_files(
            [{f"{root_dir}/testFiles/ssh_1/src/test{i}.txt": {"content": "test1234"}}]
        )

    # Use the TaskRun class to trigger the job properly
    task_runner = taskrun.TaskRun("scp-basic", "test/cfg")
    assert task_runner.run()

    # Check that the files were all transferred
    for i in range(10):
        assert os.path.exists(f"{root_dir}/testFiles/ssh_2/dest/test{i}.txt")


def test_scp_basic_pull(env_vars, setup_ssh_keys, root_dir):
    # Required files for this test:
    # ssh_1 : test/testFiles/ssh_1/src/.*\.txt

    # Create a test file
    fs.create_files(
        [{f"{root_dir}/testFiles/ssh_1/src/test.txt": {"content": "test1234"}}]
    )

    task_runner = taskrun.TaskRun("scp-basic-pull", "test/cfg")
    assert task_runner.run()


def test_scp_basic_pca_delete(env_vars, setup_ssh_keys, root_dir):
    # Required files for this test:
    # ssh_1 : test/testFiles/ssh_1/src/test1.txt
    # File will be delteted after transfer

    # Create a test file
    fs.create_files(
        [{f"{root_dir}/testFiles/ssh_1/src/test1.txt": {"content": "test1234"}}]
    )

    task_runner = taskrun.TaskRun("scp-basic-pca-delete", "test/cfg")
    assert task_runner.run()

    # Verify the file has disappeared
    assert not os.path.exists(f"{root_dir}/testFiles/ssh_1/src/test1.txt")


def test_scp_basic_pca_move(env_vars, setup_ssh_keys, root_dir):
    # Required files for this test:
    # ssh_1 : test/testFiles/ssh_1/src/test2.txt
    # File will be moved after transfer

    # Create a test file
    fs.create_files(
        [{f"{root_dir}/testFiles/ssh_1/src/test2.txt": {"content": "test1234"}}]
    )

    task_runner = taskrun.TaskRun("scp-basic-pca-move", "test/cfg")
    assert task_runner.run()

    # Verify the file has disappeared
    assert not os.path.exists(f"{root_dir}/testFiles/ssh_1/src/test2.txt")

    # Verify the file has been moved
    assert os.path.exists(f"{root_dir}/testFiles/ssh_1/archive/test2.txt")


def test_scp_source_file_conditions(env_vars, setup_ssh_keys, root_dir):
    # Required files for this test:
    # ssh_1 : test/testFiles/ssh_1/src/log\..*\.log
    # File must be >10 bytes and less than 20
    # File must be older than 60 seconds and less than 600

    # Write a 11 byte long file

    write_test_file(
        f"{root_dir}/testFiles/ssh_1/src/log.unittset.log", content="01234567890"
    )

    # This should fail, because the file is too new
    task_runner = taskrun.TaskRun("scp-source-file-conditions", "test/cfg")
    assert not task_runner.run()

    # Modify the file to be older than 1 minute and try again
    os.utime(
        f"{root_dir}/testFiles/ssh_1/src/log.unittset.log",
        (time.time() - 61, time.time() - 61),
    )

    task_runner = taskrun.TaskRun("scp-source-file-conditions", "test/cfg")
    assert task_runner.run()

    # Modify the file to be older than 10 minutes and try again
    os.utime(
        f"{root_dir}/testFiles/ssh_1/src/log.unittset.log",
        (time.time() - 601, time.time() - 601),
    )
    task_runner = taskrun.TaskRun("scp-source-file-conditions", "test/cfg")
    assert not task_runner.run()

    # Write a 9 byte long file - we need to change the age again
    write_test_file(
        f"{root_dir}/testFiles/ssh_1/src/log.unittset.log", content="012345678"
    )

    os.utime(
        f"{root_dir}/testFiles/ssh_1/src/log.unittset.log",
        (time.time() - 61, time.time() - 61),
    )

    task_runner = taskrun.TaskRun("scp-source-file-conditions", "test/cfg")
    assert not task_runner.run()

    # Write a 21 byte long file - we need to change the age again
    write_test_file(
        f"{root_dir}/testFiles/ssh_1/src/log.unittset.log",
        content="012345678901234567890",
    )
    os.utime(
        f"{root_dir}/testFiles/ssh_1/src/log.unittset.log",
        (time.time() - 61, time.time() - 61),
    )

    task_runner = taskrun.TaskRun("scp-source-file-conditions", "test/cfg")
    assert not task_runner.run()


def test_scp_file_watch(env_vars, setup_ssh_keys, root_dir):
    # Required files for this test:
    # ssh_1 : test/testFiles/ssh_1/src/.*\.log
    # File should not exist to start with

    # Create the source file
    fs.create_files(
        [{f"{root_dir}/testFiles/ssh_1/src/fileWatch.log": {"content": "01234567890"}}]
    )

    # Filewatch configured to wait 15 seconds before giving up. Expect it to fail
    task_runner = taskrun.TaskRun("scp-file-watch", "test/cfg")
    assert not task_runner.run()

    # This time, we run it again, but after 5 seconds, create the file
    # Create a thread that will run fs.create_fi:e{"s [{aft":r 5 second}}])
    t = threading.Timer(
        5,
        write_test_file,
        [f"{root_dir}/testFiles/ssh_1/src/fileWatch.txt"],
        {"content": "01234567890"},
    )
    t.start()
    logging.info("Started thread - Expect file in 5 seconds, starting task-run now...")

    task_runner = taskrun.TaskRun("scp-file-watch", "test/cfg")
    assert task_runner.run()

    # Delete the fileWatch.txt and log file
    os.remove(f"{root_dir}/testFiles/ssh_1/src/fileWatch.txt")
    os.remove(f"{root_dir}/testFiles/ssh_1/src/fileWatch.log")


def test_scp_log_watch(env_vars, setup_ssh_keys, root_dir):
    # Required files for this test:
    # ssh_1 : test/testFiles/ssh_1/src/logYYYYWatch.log
    # File should not exist to start with
    # To succeed, the file should contain the text "someText"
    year = datetime.datetime.now().year
    # Ensure the log file is removed
    log_file = f"{root_dir}/testFiles/ssh_1/src/log{year}Watch.log"
    if os.path.exists(log_file):
        os.remove(log_file)

    # Logwatch will fail if the log file doesn't exist
    task_runner = taskrun.TaskRun("scp-log-watch", "test/cfg")
    assert not task_runner.run()

    # This time, we run it again with the file created and populated. It should fail because the file doesn't contain the expected text
    # Write the file
    fs.create_files([{log_file: {"content": "NOT_THE_RIGHT_PATTERN"}}])

    task_runner = taskrun.TaskRun("scp-log-watch", "test/cfg")
    assert not task_runner.run()

    # This time we run again, but populate the file after 5 seconds
    t = threading.Timer(
        5,
        write_test_file,
        [log_file],
        {"content": "someText\n"},
    )
    t.start()
    logging.info("Started thread - Expect file in 5 seconds, starting task-run now...")
    task_runner = taskrun.TaskRun("scp-log-watch", "test/cfg")
    assert task_runner.run()


def test_scp_log_watch_tail(env_vars, setup_ssh_keys, root_dir):
    # Required files for this test:
    # ssh_1 : test/testFiles/ssh_1/src/logYYYYWatch.log
    # File should not exist to start with
    # To succeed, the file should contain the text "someText"
    year = datetime.datetime.now().year

    # Ensure the log file is removed
    if os.path.exists(f"{root_dir}/testFiles/ssh_1/src/log{year}Watch1.log"):
        os.remove(f"{root_dir}/testFiles/ssh_1/src/log{year}Watch1.log")

    # Write the matching pattern into the log, but before it runs.. This should
    # make the task fail because the pattern isn't written after the task starts
    fs.create_files(
        [
            {
                f"{root_dir}/testFiles/ssh_1/src/log{year}Watch1.log": {
                    "content": "someText\n"
                }
            }
        ]
    )
    task_runner = taskrun.TaskRun("scp-log-watch-tail", "test/cfg")
    assert not task_runner.run()

    # This time write the contents after 5 seconds
    t = threading.Timer(
        5,
        write_test_file,
        [f"{root_dir}/testFiles/ssh_1/src/log{year}Watch1.log"],
        {"content": "someText\n", "mode": "a"},
    )
    t.start()
    logging.info("Started thread - Expect file in 5 seconds, starting task-run now...")
    task_runner = taskrun.TaskRun("scp-log-watch-tail", "test/cfg")
    assert task_runner.run()


def run_task_run(task, verbose="2", config="test/cfg", noop=False):
    # We need to run the bin/task-run script to test this
    script = "python"
    args = [
        "src/opentaskpy/cli/task_run.py",
        "-t",
        task,
        "-v",
        verbose,
        "-c",
        config,
    ]

    if noop:
        args.append("--noop")

    # Run the script
    result = subprocess.run([script] + args, capture_output=True)
    # Write stdout and stderr to the console
    logging.info("\n########## STDOUT ##########")
    logging.info(result.stdout.decode("utf-8"))
    # Get the console colour for red

    logging.info("########## STDERR ##########")
    logging.info(f"{result.stderr.decode('utf-8')}")

    return {
        "returncode": result.returncode,
        "stdout": result.stdout.decode("utf-8"),
        "stderr": result.stderr.decode("utf-8"),
    }


def write_test_file(file_name, content=None, length=0, mode="w"):
    with open(file_name, mode) as f:
        if content is not None:
            f.write(content)
        else:
            f.write("a" * length)
    logging.info(f"Wrote file: {file_name}")



#!/bin/env python3
"""CLI script wrapper for handling env vars and triggering the TaskRun class."""

import argparse
import logging
import os
import sys
from datetime import datetime
from textwrap import dedent

from opentaskpy import taskrun  # type: ignore[attr-defined]
from opentaskpy.otflogging import OTF_LOG_FORMAT

CONFIG_PATH = f"{os.getcwd()}/cfg"


def main() -> None:
    """Parse args and call TaskRun class."""
    parser = argparse.ArgumentParser(
        formatter_class=argparse.RawTextHelpFormatter,
        epilog=dedent(
            """\
                                Environment Variables:

                                There are several environment variables that can be used to impact the behaviour:

                                    OTF_RUN_ID - Equivalent to using --runId argument
                                    OTF_LOG_JSON - Change the log output format to structured JSON format
                                    OTF_NO_LOG - Prevent logging to any files, will log to stdout/err only
                                    OTF_LOG_DIRECTORY - Specify a particular log directory to write log files to
                                    OTF_LOG_LEVEL - Equivalent to using -v
                                    OTF_SSH_KEY - Specify a particular SSH key to use for SSH/SFTP related transfers
                                    OTF_VARIABLES_FILE - Override the default variables file location
                                    OTF_NOOP - Equivalent to using --noop argument

                                Task Definition Overrides:

                                To override task specific values, you can use the following format in the environment variable name:
                                OTF_OVERRIDE_<TASK_TYPE>_<ATTRIBUTE>_<ATTRIBUTE>_<ATTRIBUTE>

                                e.g. OTF_OVERRIDE_TRANSFER_SOURCE_HOSTNAME

                                Case doesn't matter here. For attributes that are nested within an array, you can specify the array index

                                e.g. OTF_OVERRIDE_TRANSFER_DESTINATION_0_PROTOCOL_CREDENTIALS_USERNAME

                                """
        ),
    )

    parser.add_argument(
        "--noop",
        help=(
            "Do not attempt to run anything. Only load the config files to validate"
            " that they're OK"
        ),
        action="store_true",
        default=False,
        required=False,
    )

    parser.add_argument(
        "-t", "--taskId", help="Name of the JSON config to run", type=str, required=True
    )
    parser.add_argument(
        "-r",
        "--runId",
        help=(
            "Unique identifier to correlate logs with. e.g. if being triggered by an"
            " external scheduler"
        ),
        type=str,
        required=False,
    )
    parser.add_argument(
        "-v",
        "--verbosity",
        help="Increase verbosity:\n3 - DEBUG\n2 - VERBOSE2\n1 - VERBOSE1",
        type=int,
    )
    parser.add_argument(
        "-c", "--configDir", help="Directory containing task configurations", type=str
    )

    args = parser.parse_args()

    if args.configDir:
        global CONFIG_PATH  # pylint: disable=global-statement
        CONFIG_PATH = args.configDir

    # If given a runId, then set the environment variable
    if args.runId:
        os.environ["OTF_RUN_ID"] = args.runId

    if args.noop:
        os.environ["OTF_NOOP"] = "true"

    os.environ["OTF_LOG_RUN_PREFIX"] = datetime.now().strftime("%Y%m%d-%H%M%S.%f")[:-3]

    logging_level = logging.INFO
    if args.verbosity == 3:
        logging_level = logging.DEBUG
    elif args.verbosity == 2:
        logging_level = 11
    elif args.verbosity == 1:
        logging_level = 12

    logging.addLevelName(11, "VERBOSE2")
    logging.addLevelName(12, "VERBOSE1")

    logging.basicConfig(
        format=OTF_LOG_FORMAT,
        level=logging_level,
        handlers=[logging.StreamHandler()],
    )

    logger = logging.getLogger()
    logger.setLevel(logging_level)

    logger = logging.getLogger(__name__)
    logger.log(11, f"Log verbosity: {args.verbosity}")

    # Create the TaskRun object
    task_run_obj = taskrun.TaskRun(
        args.taskId, CONFIG_PATH, noop="OTF_NOOP" in os.environ
    )

    try:
        result = task_run_obj.run()
    except Exception as ex:  # pylint: disable=broad-exception-caught
        logger.error(f"Error running task: {ex}")
        if logger.getEffectiveLevel() <= 12:
            raise ex
        sys.exit(1)

    if not result:
        sys.exit(1)


if __name__ == "__main__":
    main()



# pylint: skip-file
import pytest

from opentaskpy.config.schemas import validate_execution_json


@pytest.fixture(scope="function")
def valid_protocol_definition():
    return {
        "name": "ssh",
        "credentials": {
            "username": "test",
        },
    }


@pytest.fixture(scope="function")
def valid_execution(valid_protocol_definition):
    return {
        "hosts": ["{{ HOST_A }}"],
        "directory": "/tmp",
        "command": "test -e test.txt",
        "protocol": valid_protocol_definition,
    }


def test_ssh_basic(valid_execution):
    json_data = {
        "type": "execution",
    }
    # Append properties from valid_execution onto json_data
    json_data.update(valid_execution)

    assert validate_execution_json(json_data)

    # Add another host
    json_data["hosts"].append("host2")

    assert validate_execution_json(json_data)

    # Remove protocol
    del json_data["protocol"]
    assert not validate_execution_json(json_data)

    # Remove all hosts
    del json_data["hosts"]
    assert not validate_execution_json(json_data)



# pylint: skip-file
import pytest

from opentaskpy.config.schemas import validate_transfer_json


@pytest.fixture(scope="function")
def valid_protocol_definition():
    return {
        "name": "ssh",
        "credentials": {
            "username": "test",
        },
    }


def test_ssh_protocol():
    json_data = {
        "type": "transfer",
        "source": {
            "hostname": "{{ HOST_A }}",
            "directory": "/tmp/testFiles/src",
            "fileRegex": ".*\\.txt",
            "protocol": {"name": "ssh"},
        },
    }

    assert not validate_transfer_json(json_data)

    # Set protocol creds
    json_data["source"]["protocol"]["credentials"] = {}
    assert not validate_transfer_json(json_data)

    json_data["source"]["protocol"]["credentials"] = {"username": "test"}
    assert validate_transfer_json(json_data)

    # Password is not supported
    json_data["source"]["protocol"]["credentials"] = {
        "username": "test",
        "password": "test",
    }
    assert not validate_transfer_json(json_data)


def test_ssh_basic(valid_protocol_definition):
    json_data = {
        "type": "transfer",
        "source": {
            "hostname": "{{ HOST_A }}",
            "directory": "/tmp/testFiles/src",
            "fileRegex": ".*\\.txt",
            "protocol": valid_protocol_definition,
        },
    }
    assert validate_transfer_json(json_data)

    # test error
    json_data["source"]["error"] = "blah"
    assert not validate_transfer_json(json_data)

    json_data["source"]["error"] = True
    assert validate_transfer_json(json_data)

    json_data["source"]["error"] = False
    assert validate_transfer_json(json_data)


def test_ssh_conditions(valid_protocol_definition):
    json_data = {
        "type": "transfer",
        "source": {
            "hostname": "{{ HOST_A }}",
            "directory": "/tmp/testFiles/src",
            "fileRegex": ".*\\.txt",
            "conditionals": {},
            "protocol": valid_protocol_definition,
        },
    }
    assert validate_transfer_json(json_data)

    # Add conditions
    for condition in ["size", "age"]:
        json_data["source"]["conditionals"][condition] = {"gt": 10}
        assert validate_transfer_json(json_data)

        json_data["source"]["conditionals"][condition] = {"lt": 10}
        assert validate_transfer_json(json_data)

        json_data["source"]["conditionals"][condition] = {"lt": 10, "gt": 5}
        assert validate_transfer_json(json_data)

    # Add an invalid condition
    json_data["source"]["conditionals"]["invalid"] = {"invalid": 10}
    assert not validate_transfer_json(json_data)


def test_ssh_filewatch(valid_protocol_definition):
    json_data = {
        "type": "transfer",
        "source": {
            "hostname": "{{ HOST_A }}",
            "directory": "/tmp/testFiles/src",
            "fileRegex": ".*\\.txt",
            "fileWatch": {},
            "protocol": valid_protocol_definition,
        },
    }
    assert validate_transfer_json(json_data)

    # Add fileWatch config
    json_data["source"]["fileWatch"] = {"timeout": 10}
    assert validate_transfer_json(json_data)

    # Add a file and directory
    json_data["source"]["fileWatch"]["fileRegex"] = ".*"
    json_data["source"]["fileWatch"]["directory"] = "/tmp"
    assert validate_transfer_json(json_data)

    # Add watchOnly, and remove the fileRegex and directory
    json_data["source"]["fileWatch"]["watchOnly"] = True
    del json_data["source"]["fileRegex"]
    del json_data["source"]["directory"]
    assert not validate_transfer_json(json_data)

    # Add an invalid property
    json_data["source"]["fileWatch"]["invalid"] = "invalid"
    assert not validate_transfer_json(json_data)


def test_ssh_logwatch(valid_protocol_definition):
    json_data = {
        "type": "transfer",
        "source": {
            "hostname": "{{ HOST_A }}",
            "directory": "/tmp/testFiles/src",
            "fileRegex": ".*\\.txt",
            "logWatch": {},
            "protocol": valid_protocol_definition,
        },
    }
    assert not validate_transfer_json(json_data)

    json_data = {
        "type": "transfer",
        "source": {
            "hostname": "{{ HOST_A }}",
            "directory": "/tmp/testFiles/src",
            "fileRegex": ".*\\.txt",
            "logWatch": {
                "log": "test.log",
                "directory": "/tmp",
                "contentRegex": ".*",
            },
            "protocol": valid_protocol_definition,
        },
    }
    assert validate_transfer_json(json_data)

    # Add tail
    json_data["source"]["logWatch"]["tail"] = True
    assert validate_transfer_json(json_data)

    # Add an invalid property
    json_data["source"]["logWatch"]["invalid"] = "invalid"
    assert not validate_transfer_json(json_data)


def test_ssh_pca(valid_protocol_definition):
    json_data = {
        "type": "transfer",
        "source": {
            "hostname": "{{ HOST_A }}",
            "directory": "/tmp/testFiles/src",
            "fileRegex": ".*\\.txt",
            "postCopyAction": {},
            "protocol": valid_protocol_definition,
        },
    }
    assert not validate_transfer_json(json_data)

    # PCA delete
    json_data["source"]["postCopyAction"]["action"] = "delete"
    assert validate_transfer_json(json_data)

    # PCA move
    json_data["source"]["postCopyAction"]["action"] = "move"
    assert not validate_transfer_json(json_data)

    json_data["source"]["postCopyAction"]["destination"] = "/tmp/testFiles/dest"
    assert validate_transfer_json(json_data)

    # PCA copy
    json_data["source"]["postCopyAction"]["action"] = "rename"
    assert not validate_transfer_json(json_data)

    json_data["source"]["postCopyAction"]["sub"] = "test"
    json_data["source"]["postCopyAction"]["pattern"] = "test"
    assert validate_transfer_json(json_data)



"""Abstract classes for remote handlers."""

from abc import ABC, abstractmethod


class RemoteHandler(ABC):
    """Parent class for remote handlers."""

    def __init__(self, spec: dict):
        """Initialise the handler.

        Args:
            spec (dict): The spec for the handler.
        """
        self.spec = spec


class RemoteTransferHandler(RemoteHandler):
    """Abstract class for remote transfer handlers."""

    def __init__(self, spec: dict, remote_spec: dict | None = None):
        """Initialise the handler.

        Args:
            spec (dict): The spec for the transfer.
            remote_spec (dict, optional): The remote spec for the transfer. Defaults to None.
        """
        self.remote_spec = remote_spec
        # Call super with the spec
        super().__init__(spec)

    @abstractmethod
    def supports_direct_transfer(self) -> bool:
        """Check if the remote handler supports direct transfers."""

    @abstractmethod
    def list_files(
        self, directory: str | None = None, file_pattern: str | None = None
    ) -> dict:
        """Generate a list of files to transfer."""

    @abstractmethod
    def transfer_files(
        self,
        files: list[str],
        remote_spec: dict,
        dest_remote_handler: dict | None = None,
    ) -> int:
        """Transfer files to the remote location.

        Args:
            files (list[str]): The files to transfer.
            remote_spec (dict): The remote spec for the transfer.
            dest_remote_handler (RemoteTransferHandler, optional): The remote handler for the destination. Defaults to None.

        Returns:
            int: The result of the transfer. 0 for success, 1 for failure.
        """

    @abstractmethod
    def push_files_from_worker(self, local_staging_directory: str) -> int:
        """Push files from the worker to the remote location.

        This is used when files have been either generated on the worker, or copied
        as a proxy transfer onto the worker.

        Args:
            local_staging_directory (str): The local staging directory.

        Returns:
            int: The result of the transfer. 0 for success, 1 for failure.
        """

    @abstractmethod
    def pull_files_to_worker(
        self, files: list[str], local_staging_directory: str
    ) -> int:
        """Pull files from the remote location to the worker.

        Args:
            files (list[str]): The files to pull.
            local_staging_directory (str): The local staging directory.

        Returns:
            int: The result of the transfer. 0 for success, 1 for failure.
        """

    @abstractmethod
    def pull_files(self, files: list[str]) -> int:
        """Pull files from the remote location to the destination system.

        Used when a direct push cannot be done, and the files need to be pulled instead.

        Args:
            files (list[str]): The files to pull.

        Returns:
            int: The result of the transfer. 0 for success, 1 for failure.
        """

    @abstractmethod
    def move_files_to_final_location(self, files: dict) -> int:
        """Move files to their final location.

        Once dropped on the destination system, the files may need to be moved to their
        final location.

        Args:
            files (dict): The files to move.

        Returns:
            int: The result of the transfer. 0 for success, 1 for failure.
        """

    @abstractmethod
    def handle_post_copy_action(self, files: list[str]) -> int:
        """Handle any post copy actions.

        Post Copy Actions (PCA) are actions that need to be performed after the files
        have been copied to the destination system and are in their final location. The
        PCA acts on the source files, and will typically delete them, move them, or
        rename them.

        Args:
            files (list[str]): The files to act on.

        Returns:
            int: The result of the transfer. 0 for success, 1 for failure.
        """

    def tidy(self) -> None:
        """Tidy up after the transfer, if necessary. Otherwise do nothing."""


class RemoteExecutionHandler(RemoteHandler):
    """Abstract class for remote execution handlers."""

    @abstractmethod
    def execute(self) -> bool:
        """Execute the command.

        Returns:
            bool: True if the command was successful, False otherwise.
        """



# pylint: skip-file
import os

import pytest
from pytest_shell import fs

from opentaskpy import exceptions
from opentaskpy.taskhandlers import execution
from tests.fixtures.ssh_clients import *  # noqa: F403

os.environ["OTF_NO_LOG"] = "1"
os.environ["OTF_LOG_LEVEL"] = "DEBUG"

# Create a task definition
touch_task_definition = {
    "type": "execution",
    "hosts": ["172.16.0.11", "172.16.0.12"],
    "directory": "/tmp",
    "command": "touch /tmp/testFiles/dest/execution.txt",
    "protocol": {"name": "ssh", "credentials": {"username": "application"}},
}

fail_task_definition = {
    "type": "execution",
    "hosts": ["172.16.0.11", "172.16.0.12"],
    "directory": "/tmp",
    "command": "test -e /tmp/testFiles/src/execution.test.fail.txt",
    "protocol": {"name": "ssh", "credentials": {"username": "application"}},
}

fail_host_task_definition = {
    "type": "execution",
    "hosts": ["172.16.0.11", "172.16.255.12"],
    "directory": "/tmp",
    "command": "touch /tmp/testFiles/dest/execution.invalidhost.txt",
    "protocol": {"name": "ssh", "credentials": {"username": "application"}},
}

fail_invalid_protocol_task_definition = {
    "type": "execution",
    "hosts": ["172.16.0.11", "172.16.255.12"],
    "directory": "/tmp",
    "command": "touch /tmp/testFiles/dest/execution.invalidhost.txt",
    "protocol": {"name": "rubbish"},
}


def test_invalid_protocol():
    execution_obj = execution.Execution(
        None, "invalid-protocol", fail_invalid_protocol_task_definition
    )
    # Expect a UnknownProtocolError exception
    with pytest.raises(exceptions.UnknownProtocolError):
        execution_obj._set_remote_handlers()


def test_basic_execution(setup_ssh_keys, root_dir):
    execution_obj = execution.Execution(None, "df-basic", touch_task_definition)
    execution_obj._set_remote_handlers()

    # Ensure no test files exist already, if so delete them
    if os.path.exists(f"{root_dir}/testFiles/ssh_1/dest/execution.txt"):
        os.remove(f"{root_dir}/testFiles/ssh_1/dest/execution.txt")

    if os.path.exists(f"{root_dir}/testFiles/ssh_2/dest/execution.txt"):
        os.remove(f"{root_dir}/testFiles/ssh_2/dest/execution.txt")

    # Validate some things were set as expected
    assert execution_obj.remote_handlers[0].__class__.__name__ == "SSHExecution"

    assert execution_obj.remote_handlers[1].__class__.__name__ == "SSHExecution"

    # Run the execution and expect a true status
    assert execution_obj.run()

    # Check the destination file exists on both hosts
    assert os.path.exists(f"{root_dir}/testFiles/ssh_1/dest/execution.txt")
    assert os.path.exists(f"{root_dir}/testFiles/ssh_2/dest/execution.txt")


def test_basic_execution_cmd_failure(setup_ssh_keys, root_dir):
    # Write a test file to the source directory
    fs.create_files(
        [
            {
                f"{root_dir}/testFiles/ssh_1/src/execution.test.fail.txt": {
                    "content": "test1234"
                }
            }
        ]
    )

    execution_obj = execution.Execution(None, "task-fail", fail_task_definition)
    execution_obj._set_remote_handlers()

    # Validate some things were set as expected
    assert execution_obj.remote_handlers[0].__class__.__name__ == "SSHExecution"

    assert execution_obj.remote_handlers[1].__class__.__name__ == "SSHExecution"

    # Run the execution and expect a failure

    assert not execution_obj.run()


def test_basic_execution_invalid_host(setup_ssh_keys, root_dir):
    execution_obj = execution.Execution(None, "task-fail", fail_host_task_definition)
    execution_obj._set_remote_handlers()
    # Remove test files if they exist
    if os.path.exists(f"{root_dir}/testFiles/ssh_1/dest/execution.invalidhost.txt"):
        os.remove(f"{root_dir}/testFiles/ssh_1/dest/execution.invalidhost.txt")

    # Validate some things were set as expected
    assert execution_obj.remote_handlers[0].__class__.__name__ == "SSHExecution"

    assert execution_obj.remote_handlers[1].__class__.__name__ == "SSHExecution"

    assert not execution_obj.run()

    # But the remote file should still have been created on the valid host
    assert os.path.exists(f"{root_dir}/testFiles/ssh_1/dest/execution.invalidhost.txt")

    assert not os.path.exists(
        f"{root_dir}/testFiles/ssh_2/dest/execution.invalidhost.txt"
    )



"""Batch task handler class."""

import threading
import time
from concurrent.futures import ThreadPoolExecutor, wait

import opentaskpy.otflogging
from opentaskpy.config.loader import ConfigLoader
from opentaskpy.exceptions import InvalidConfigError
from opentaskpy.taskhandlers.execution import Execution
from opentaskpy.taskhandlers.taskhandler import TaskHandler
from opentaskpy.taskhandlers.transfer import Transfer

DEFAULT_TASK_TIMEOUT = 300
DEFAULT_TASK_CONTINUE_ON_FAIL = False
DEFAULT_TASK_RETRY_ON_RERUN = False
DEFAULT_TASK_EXIT_CODE = 0
TASK_TYPE = "B"
BATCH_TASK_LOG_MARKER = "__OTF_BATCH_TASK_MARKER__"


class Batch(TaskHandler):
    """Batch task handler class."""

    overall_result = False

    def __init__(
        self,
        global_config: dict,
        task_id: str,
        batch_definition: dict,
        config_loader: ConfigLoader,
    ):
        """Initialise the batch task handler.

        Args:
            global_config (dict): The global configuration for the task.
            task_id (str): The ID of the task.
            batch_definition (dict): The definition of the batch task.
            config_loader (ConfigLoader): The config loader to use to load
            task definitions within the batch.
        """
        self.task_id = task_id
        self.batch_definition = batch_definition
        self.config_loader = config_loader
        self.tasks = {}
        self.task_order_tree = {}

        super().__init__(global_config)

        self.logger = opentaskpy.otflogging.init_logging(
            __name__, self.task_id, TASK_TYPE
        )

        # We need to get the latest log file (if there is one), to determine
        # where to start from (if we are resuming a batch)
        previous_log_file = opentaskpy.otflogging.get_latest_log_file(
            self.task_id, TASK_TYPE
        )

        self.logger.info(
            f"Found previous log file: {previous_log_file}"
            if previous_log_file
            else "No previous log file found"
        )

        previous_status = {}
        if previous_log_file:
            self.logger.info("Parsing previous log file for log marks")

            # Parse the previous log file to determine where we left off
            with open(previous_log_file, encoding="utf-8") as f:
                for line in f:
                    if BATCH_TASK_LOG_MARKER in line:
                        log_mark = line.split(f"{BATCH_TASK_LOG_MARKER}: ")[1].strip()
                        self.logger.info(f"Found log mark: {log_mark}")
                        # Mark the task with the appropriate status
                        # Determine which order id this is for
                        mark_metadata = log_mark.split("::")
                        order_id = int(mark_metadata[1])
                        task_status = mark_metadata[4]

                        previous_status[order_id] = task_status

        # Order the batch_definitions by order_id
        self.batch_definition["tasks"] = sorted(
            self.batch_definition["tasks"], key=lambda k: k["order_id"]  # type: ignore[no-any-return]
        )

        # Parse the batch definition and create the appropriate tasks to run
        # in the correct order, based on the dependencies specified
        for task in self.batch_definition["tasks"]:
            # Load the definition for the task
            task_definition = self.config_loader.load_task_definition(task["task_id"])
            order_id = task["order_id"]

            # Add it to the list of tasks
            self.tasks[order_id] = task_definition

            # Set the timeout for the task
            timeout = task.get("timeout", DEFAULT_TASK_TIMEOUT)

            # Same for continue on fail
            continue_on_fail = None
            if "continue_on_fail" in task:
                continue_on_fail = task["continue_on_fail"]
            else:
                continue_on_fail = DEFAULT_TASK_CONTINUE_ON_FAIL

            # Same for retry on rerun
            retry_on_rerun = None
            retry_on_rerun = task.get("retry_on_rerun", DEFAULT_TASK_RETRY_ON_RERUN)

            # Set the status of the task
            status = "NOT_STARTED"
            if order_id in previous_status:
                # Check whether it should be rerun if it worked last time
                if retry_on_rerun:
                    if previous_status[order_id] == "COMPLETED":
                        # Log something to make it clear that we are rerunning
                        self.logger.info(
                            f"Task {task_id} succeeded last time, but is marked to be"
                            " rerun"
                        )
                    status = "NOT_STARTED"
                elif previous_status[order_id] == "COMPLETED":
                    self.logger.info(
                        f"Task {task_id} succeeded last time, and is not marked to be"
                        " rerun, so marking as complete"
                    )
                    # Output the log mark
                    self._log_task_result(
                        "COMPLETED",
                        str(order_id),
                        task["task_id"],
                    )
                    status = "COMPLETED"

            # Create the task handlers for anything we intend on running
            task_handler = None
            if status == "NOT_STARTED":
                # Create the appropriate task handler based on the task type
                if task_definition["type"] == "execution":
                    task_handler = Execution(
                        global_config, task["task_id"], task_definition
                    )
                elif task_definition["type"] == "transfer":
                    task_handler = Transfer(
                        global_config, task["task_id"], task_definition
                    )
                elif task_definition["type"] == "batch":
                    task_handler = Batch(
                        global_config,
                        task["task_id"],
                        task_definition,
                        self.config_loader,
                    )
                else:
                    raise InvalidConfigError("Unknown task type")

            self.task_order_tree[order_id] = {
                "task_id": task["task_id"],
                "batch_task_spec": task,
                "task": task_definition,
                "task_handler": task_handler,
                "timeout": timeout,
                "continue_on_fail": continue_on_fail,
                "retry_on_rerun": retry_on_rerun,
                "status": status,
                "result": None,
            }

        # Debug to show the structure of the tree
        self.logger.debug(f"Task order tree: {self.task_order_tree}")

    def _set_remote_handlers(self) -> None:
        pass

    def return_result(
        self,
        status: int,
        message: str | None = None,
        exception: type[Exception] | None = None,
    ) -> bool:
        """Return the result of the task run.

        Args:
            status (int): The status code to return.
            message (str, optional): The message to return. Defaults to None.
            exception (Exception, optional): The exception to return. Defaults to None.

        Returns:
            bool: The result of the task run.
        """
        return super().return_result(status, message, exception)  # type: ignore[no-any-return]

    def run(self, kill_event: threading.Event | None = None) -> bool:
        """Run the batch.

        Args:
            kill_event (threading.Event, optional): An event to kill the batch. Defaults to None.

        Returns:
            bool: True if the batch completed successfully, False otherwise.
        """
        self.logger.info("Running batch")

        # This is where we could potentially be waiting for a while. So,
        # for each task handler, we should spawn a new thread to run
        # the command. Then we can wait for all threads to complete.

        ex = None
        while True:
            # Loop through every task in the tree
            for order_id, batch_task in self.task_order_tree.items():
                task = batch_task["task"]
                logged = False
                if "logged_status" in batch_task:
                    logged = batch_task["logged_status"]
                else:
                    batch_task["logged_status"] = False

                self.logger.log(
                    12, f"Checking task {order_id} ({batch_task['task_id']})"
                )

                # Check if there are dependencies for this task that have not yet completed, if so then we skip it
                if "dependencies" in batch_task["batch_task_spec"]:
                    all_dependencies_complete = True
                    for dependency in batch_task["batch_task_spec"]["dependencies"]:
                        if self.task_order_tree[dependency]["status"] != "COMPLETED":
                            self.logger.log(
                                12,
                                (
                                    "Skipping task"
                                    f" {order_id} ({batch_task['task_id']}) as"
                                    f" dependency {dependency} has not completed"
                                ),
                            )
                            all_dependencies_complete = False
                            continue
                        self.logger.info(
                            "All dependencies for task"
                            f" {order_id} ({batch_task['task_id']}) have completed"
                        )
                    if not all_dependencies_complete:
                        continue

                # Check if the task has already been triggered
                if batch_task["status"] == "NOT_STARTED":
                    # Create a new thread to run the task. Pass an event into the thread so we can handle timeouts
                    batch_task["executing_thread"] = None
                    e = threading.Event()
                    thread = threading.Thread(
                        target=self.task_runner,
                        args=(batch_task, e),
                        name=f"{batch_task['task_id']}_parent",
                    )
                    # Start the thread
                    thread.start()

                    self.logger.info(
                        f"Spawned thread for task {order_id} ({batch_task['task_id']})"
                    )
                    batch_task["status"] = "RUNNING"
                    batch_task["start_time"] = time.time()
                    batch_task["thread"] = thread
                    batch_task["kill_event"] = e

                    # Sleep 1 second to allow the thread to start
                    time.sleep(1)

                # Check if the task has completed
                if batch_task["status"] == "RUNNING":
                    # Check if the task has timed out
                    if (
                        time.time() - batch_task["start_time"] > batch_task["timeout"]
                        and batch_task["thread"].is_alive()
                    ):
                        self.logger.error(
                            f"Task {order_id} ({batch_task['task_id']}) has timed out"
                        )
                        logged = True
                        batch_task["status"] = "TIMED_OUT"
                        # Send event to the thread to kill it
                        self.logger.debug(
                            f"Sending kill event to task {order_id} ({batch_task['task_id']})"
                        )
                        batch_task["kill_event"].set()
                        # Wait for the thread to return
                        batch_task["thread"].join()
                        self.logger.debug(
                            f"Task {order_id} ({batch_task['task_id']}) has been killed"
                        )
                        batch_task["result"] = False

                    # Check whether the thread is actually still running.
                    # If it has died uncleanly, then we need to set the appropriate statuses for it
                    if (
                        not batch_task["thread"].is_alive()
                        and batch_task["status"] != "TIMED_OUT"
                    ):
                        self.logger.error(
                            f"Task {order_id} ({batch_task['task_id']}) has failed"
                        )
                        logged = True
                        batch_task["status"] = "FAILED"
                        batch_task["result"] = False

                if batch_task["status"] == "COMPLETED" and "thread" in batch_task:
                    batch_task["thread"].join()
                    if not logged:
                        self.logger.info(
                            f"Task {order_id} ({batch_task['task_id']}) has completed"
                        )
                        logged = True
                    else:
                        self.logger.log(
                            12,
                            f"Task {order_id} ({batch_task['task_id']}) has completed",
                        )

                # Add a generic message to show the task failed
                if batch_task["status"] == "FAILED":
                    if not logged:
                        self.logger.error(
                            f"Task {order_id} ({batch_task['task_id']}) has failed"
                        )
                        logged = True
                    else:
                        self.logger.log(
                            12, f"Task {order_id} ({batch_task['task_id']}) has failed"
                        )

                # Handle instances where we timed out or failed, and we should continue on fail
                if (
                    batch_task["status"] in ["TIMED_OUT", "FAILED"]
                    and batch_task["continue_on_fail"]
                ):
                    self.logger.info(
                        f"Task {order_id} ({batch_task['task_id']}) has failed, but"
                        " continuing on fail"
                    )
                    batch_task["status"] = "COMPLETED"
                    batch_task["result"] = False
                    logged = True

                batch_task["logged_status"] = logged

            # Check if there are any tasks that are still in RUNNING state, if not then we are done
            running_tasks = [
                task
                for task in self.task_order_tree.values()
                if task["status"] == "RUNNING"
            ]
            if len(running_tasks) == 0:
                break

            # Sleep 5 seconds before checking again
            time.sleep(5)

            # Check if we have been asked to kill the batch
            if kill_event and kill_event.is_set():
                self.logger.info("Kill event received, stopping threads")
                for _order_id, batch_task in self.task_order_tree.items():
                    if batch_task["status"] == "RUNNING":
                        batch_task["kill_event"].set()
                        # Give the thread 2 seconds to stop, otherwise we kill it
                        batch_task["thread"].join(2)
                        if batch_task["thread"].is_alive():
                            batch_task["thread"].cancel()

                        batch_task["status"] = "KILLED"

        # Check if any tasks have failed
        failed_tasks = [
            task
            for task__ in self.task_order_tree.values()
            if task__["status"] != "COMPLETED" or task__["result"] is False
        ]
        if len(failed_tasks) == 0:
            self.overall_result = True

        self.logger.info(f"Batch completed with result {self.overall_result}")

        if self.overall_result:
            return self.return_result(0, "All batch tasks completed successfully")

        return self.return_result(1, "Batch failed", ex)

    def task_runner(self, batch_task: dict, event: threading.Event) -> None:
        """Run the task in a separate thread.

        Args:
            batch_task (dict): The task to run
            event (threading.Event): The event to use to kill the thread
        """
        self.logger.info(f"Running task {batch_task['task_id']}")
        with ThreadPoolExecutor(
            max_workers=1, thread_name_prefix=batch_task["task_id"]
        ) as executor:
            while True:
                if batch_task["executing_thread"] is None:
                    self.logger.log(
                        12,
                        f"Spawning task handler for {batch_task['task_id']} with timeout of {batch_task['timeout']}",
                    )
                    # Spawn the task as it's own thread
                    batch_task["executing_thread"] = [
                        executor.submit(
                            self._execute, batch_task["task_handler"], event
                        )
                    ]
                else:
                    self.logger.log(
                        12, f"Checking task handler for {batch_task['task_id']}"
                    )
                    # Check to see if the thread is still alive
                    if (
                        not batch_task["executing_thread"][0].running()
                        and batch_task["status"] != "TIMED_OUT"
                    ):
                        # Get the returncode from the thread

                        batch_task["result"] = batch_task["executing_thread"][
                            0
                        ].result()

                        if batch_task["result"]:
                            batch_task["status"] = "COMPLETED"
                        else:
                            batch_task["status"] = "FAILED"
                        self._log_task_result(
                            batch_task["status"],
                            batch_task["batch_task_spec"]["order_id"],
                            batch_task["task_id"],
                        )
                        break

                    # Check if we have been asked to kill the thread
                    if event.is_set():
                        self.logger.log(
                            12, f"Killing task handler for {batch_task['task_id']}"
                        )
                        # Kill the thread and all it's child processes
                        batch_task["executing_thread"][0].cancel()
                        executor.shutdown(wait=False)
                        self._log_task_result(
                            "FAILED",
                            batch_task["batch_task_spec"]["order_id"],
                            batch_task["task_id"],
                        )

                        break

                    # Wait for the thread to complete
                    self.logger.log(
                        12, f"Waiting for task handler for {batch_task['task_id']}"
                    )
                    wait(batch_task["executing_thread"], timeout=2)

            batch_task["end_time"] = time.time()

    def _log_task_result(self, status: str, order_id: str, task_id: str) -> None:
        self.logger.info(
            f"{BATCH_TASK_LOG_MARKER}: ORDER_ID::{order_id}::TASK::{task_id}::{status}"
        )

    def _execute(
        self, task_handler: TaskHandler, event: threading.Event | None = None
    ) -> bool:  #
        result = False
        try:
            result = task_handler.run(kill_event=event)
        except Exception as ex:  # pylint: disable=broad-exception-caught
            task_handler.return_result(1, str(ex), ex)
            self.logger.error(f"[{task_handler.task_id}] Failed to run task")
            # Log the call stack
            self.logger.exception(ex)
            result = False

        self.logger.info(f"[{task_handler.task_id}] Returned {result}")
        return result



# pylint: skip-file
import os
import random
import shutil

import pytest
from pytest_shell import fs

import opentaskpy.otflogging
from opentaskpy.config.loader import ConfigLoader

# from opentaskpy.taskhandlers.batch import Batch
from opentaskpy.taskhandlers import batch, execution, transfer
from tests.file_helper import *  # noqa: F403
from tests.fixtures.ssh_clients import *  # noqa: F403

os.environ["OTF_NO_LOG"] = "1"
os.environ["OTF_LOG_LEVEL"] = "DEBUG"


# Create a task definition
basic_batch_definition = {
    "type": "batch",
    "tasks": [
        {
            "order_id": 1,
            "task_id": "touch",
        }
    ],
}

timeout_batch_definition = {
    "type": "batch",
    "tasks": [
        {
            "order_id": 1,
            "task_id": "sleep-300",
            "timeout": 10,
        },
        {
            "order_id": 2,
            "task_id": "sleep-300-local",
            "timeout": 10,
        },
    ],
}

timeout_batch_transfer_definition = {
    "type": "batch",
    "tasks": [
        {
            "order_id": 1,
            "task_id": "filewatch-300",
            "timeout": 10,
        },
        {
            "order_id": 2,
            "task_id": "filewatch-local-300",
            "timeout": 10,
        },
    ],
}

parallel_batch_definition = {
    "type": "batch",
    "tasks": [
        {
            "order_id": 1,
            "task_id": "touch",
        },
        {
            "order_id": 2,
            "task_id": "scp-basic",
        },
    ],
}

parallel_batch_with_single_failure_definition = {
    "type": "batch",
    "tasks": [
        {
            "order_id": 1,
            "task_id": "touch",
        },
        {
            "order_id": 2,
            "task_id": "fail-command",
        },
    ],
}

parallel_batch_with_single_failure_with_retry_definition = {
    "type": "batch",
    "tasks": [
        {
            "order_id": 1,
            "task_id": "touch",
            "retry_on_rerun": True,
        },
        {
            "order_id": 2,
            "task_id": "fail-command",
        },
    ],
}

parallel_batch_many_definition = {
    "type": "batch",
    "tasks": [
        {
            "order_id": 1,
            "task_id": "sleep-5",
        },
        {
            "order_id": 2,
            "task_id": "sleep-5",
        },
        {
            "order_id": 3,
            "task_id": "sleep-5",
        },
        {
            "order_id": 4,
            "task_id": "sleep-5",
        },
        {
            "order_id": 5,
            "task_id": "sleep-5",
        },
        {
            "order_id": 6,
            "task_id": "sleep-5",
        },
        {
            "order_id": 7,
            "task_id": "sleep-5",
        },
        {
            "order_id": 8,
            "task_id": "sleep-5",
        },
        {
            "order_id": 9,
            "task_id": "sleep-5",
        },
        {
            "order_id": 10,
            "task_id": "sleep-5",
        },
    ],
}

dependent_batch_definition = {
    "type": "batch",
    "tasks": [
        {
            "order_id": 1,
            "task_id": "touch",
        },
        {
            "order_id": 2,
            "task_id": "scp-basic",
            "dependencies": [
                1,
            ],
        },
    ],
}

dependent_batch_continue_on_fail_definition = {
    "type": "batch",
    "tasks": [
        {
            "order_id": 1,
            "task_id": "fail-command",
            "continue_on_fail": True,
        },
        {
            "order_id": 2,
            "task_id": "scp-basic",
            "dependencies": [
                1,
            ],
        },
    ],
}

touch_task_definition = {
    "type": "execution",
    "hosts": ["172.16.0.11", "172.16.0.12"],
    "username": "application",
    "directory": "/tmp",
    "command": "touch /tmp/testFiles/dest/execution.txt",
    "protocol": {"name": "ssh", "credentials": {"username": "application"}},
}

fail_batch_definition = {
    "type": "batch",
    "tasks": [
        {
            "order_id": 1,
            "task_id": "non-existent",
        }
    ],
}

# Create a variable with a random number
RANDOM = random.randint(10000, 99999)


def test_basic_batch(setup_ssh_keys, env_vars, root_dir, clear_logs):
    # Create a test file
    fs.create_files(
        [{f"{root_dir}/testFiles/ssh_1/src/test.txt": {"content": "test1234"}}]
    )

    # We need a config loader object, so that the batch class can load in the configs for
    # the sub tasks
    config_loader = ConfigLoader("test/cfg")
    batch_obj = batch.Batch(
        None, f"basic-{RANDOM}", basic_batch_definition, config_loader
    )

    # Check that the batch_obj contains an execution type task
    # batch_obj.tasks[0] should be an instance of a execution task handler class
    assert isinstance(
        batch_obj.task_order_tree[1]["task_handler"],
        execution.Execution,
    )

    # Run the batch and expect a true status
    assert batch_obj.run()

    # Check the execution task was run
    assert os.path.exists(f"{root_dir}/testFiles/ssh_1/src/touchedFile.txt")


def test_batch_parallel(setup_ssh_keys, env_vars, root_dir, clear_logs):
    # Create a test file
    fs.create_files(
        [{f"{root_dir}/testFiles/ssh_1/src/test.txt": {"content": "test1234"}}]
    )

    # We need a config loader object, so that the batch class can load in the configs for
    # the sub tasks
    config_loader = ConfigLoader("test/cfg")
    batch_obj = batch.Batch(
        None, f"parallel-{RANDOM}", parallel_batch_definition, config_loader
    )

    # Check that the batch_obj contains an execution type task
    # batch_obj.tasks[0] should be an instance of a execution task handler class
    assert isinstance(batch_obj.task_order_tree[1]["task_handler"], execution.Execution)
    assert isinstance(batch_obj.task_order_tree[2]["task_handler"], transfer.Transfer)

    # Run the batch and expect a true status
    assert batch_obj.run()


def test_batch_dependencies(root_dir, setup_ssh_keys, env_vars, clear_logs):
    # Create a test file
    fs.create_files(
        [{f"{root_dir}/testFiles/ssh_1/src/test.txt": {"content": "test1234"}}]
    )

    # We need a config loader object, so that the batch class can load in the configs for
    # the sub tasks
    config_loader = ConfigLoader("test/cfg")
    batch_obj = batch.Batch(
        None, f"dependencies-1-{RANDOM}", dependent_batch_definition, config_loader
    )

    # Check that the batch_obj contains an execution type task
    # batch_obj.tasks[0] should be an instance of a execution task handler class
    assert isinstance(batch_obj.task_order_tree[1]["task_handler"], execution.Execution)
    assert isinstance(batch_obj.task_order_tree[2]["task_handler"], transfer.Transfer)

    # Run the batch and expect a true status
    assert batch_obj.run()


def test_batch_invalid_task_id(root_dir, setup_ssh_keys, env_vars, clear_logs):
    # Create a test file
    fs.create_files(
        [{f"{root_dir}/testFiles/ssh_1/src/test.txt": {"content": "test1234"}}]
    )

    # We need a config loader object, so that the batch class can load in the configs for
    # the sub tasks
    config_loader = ConfigLoader("test/cfg")
    # Expect a FileNotFoundError as the task_id is non-existent
    with pytest.raises(FileNotFoundError):
        batch.Batch(None, f"fail-{RANDOM}", fail_batch_definition, config_loader)


def test_batch_execution_timeout(setup_ssh_keys, env_vars, root_dir, clear_logs):
    # Set a log file prefix for easy identification
    # Get a random number

    os.environ["OTF_LOG_RUN_PREFIX"] = f"testbatch_timeout_{RANDOM}"

    config_loader = ConfigLoader("test/cfg")
    batch_obj = batch.Batch(None, "timeout", timeout_batch_definition, config_loader)
    assert not batch_obj.run()

    # Validate that a log has been created with the correct status
    # Use the logging module to get the right log file name
    log_file_name_batch = opentaskpy.otflogging._define_log_file_name("timeout", "B")
    log_file_name_task = opentaskpy.otflogging._define_log_file_name("sleep-300", "E")
    log_file_name_task_local = opentaskpy.otflogging._define_log_file_name(
        "sleep-300-local", "E"
    )

    # Check that both exist, but renamed with _failed
    assert os.path.exists(log_file_name_batch.replace("_running", "_failed"))
    assert os.path.exists(log_file_name_task.replace("_running", "_failed"))
    assert os.path.exists(log_file_name_task_local.replace("_running", "_failed"))

    # Check the contents of the batch log, and verify that it states each task has timed
    # out (and not that it has errored for another reason)
    with open(
        log_file_name_batch.replace("_running", "_failed"), encoding="utf-8"
    ) as f:
        batch_log = f.read()
        assert "Task 1 (sleep-300) has timed out" in batch_log
        assert "Task 2 (sleep-300-local) has timed out" in batch_log


def test_batch_transfer_timeout(setup_ssh_keys, env_vars, root_dir, clear_logs):
    # Set a log file prefix for easy identification
    # Get a random number

    os.environ["OTF_LOG_RUN_PREFIX"] = f"testbatch_transfer_timeout_{RANDOM}"

    config_loader = ConfigLoader("test/cfg")
    batch_obj = batch.Batch(
        None, "transfer_timeout", timeout_batch_transfer_definition, config_loader
    )
    assert not batch_obj.run()

    # Validate that a log has been created with the correct status
    # Use the logging module to get the right log file name
    log_file_name_batch = opentaskpy.otflogging._define_log_file_name(
        "transfer_timeout", "B"
    )
    log_file_name_task = opentaskpy.otflogging._define_log_file_name(
        "filewatch-300", "T"
    )
    log_file_name_task_local = opentaskpy.otflogging._define_log_file_name(
        "filewatch-local-300", "T"
    )

    # Check that both exist, but renamed with _failed
    assert os.path.exists(log_file_name_batch.replace("_running", "_failed"))
    assert os.path.exists(log_file_name_task.replace("_running", "_failed"))
    assert os.path.exists(log_file_name_task_local.replace("_running", "_failed"))

    # Check the contents of the batch log, and verify that it states each task has timed
    # out (and not that it has errored for another reason)
    with open(
        log_file_name_batch.replace("_running", "_failed"), encoding="utf-8"
    ) as f:
        batch_log = f.read()
        assert "Task 1 (filewatch-300) has timed out" in batch_log
        assert "Task 2 (filewatch-local-300) has timed out" in batch_log


def test_batch_parallel_single_success(setup_ssh_keys, env_vars, root_dir, clear_logs):
    # Forcing a prefix makes it easy to identify log files, as well as
    # ensuring that any rerun logic doesn't get hit
    os.environ["OTF_LOG_RUN_PREFIX"] = f"testbatch_timeout_{RANDOM}"

    task_id = f"parallel-single-failure-{RANDOM}"

    config_loader = ConfigLoader("test/cfg")
    batch_obj = batch.Batch(
        None,
        task_id,
        parallel_batch_with_single_failure_definition,
        config_loader,
    )
    # Run and expect a false status
    assert not batch_obj.run()

    # Validate that a log has been created with the correct status
    # Use the logging module to get the right log file name
    log_file_name_batch = opentaskpy.otflogging._define_log_file_name(task_id, "B")
    log_file_name_touch_task = opentaskpy.otflogging._define_log_file_name("touch", "E")
    log_file_name_failed_task = opentaskpy.otflogging._define_log_file_name(
        "fail-command", "E"
    )

    # Check that all exist, with the right status
    assert os.path.exists(log_file_name_batch.replace("_running", "_failed"))
    assert os.path.exists(log_file_name_touch_task.replace("_running", ""))
    assert os.path.exists(log_file_name_failed_task.replace("_running", "_failed"))


def test_batch_resume_after_failure(setup_ssh_keys, env_vars, root_dir, clear_logs):
    task_id = f"parallel-single-failure-1-{RANDOM}"
    # Ensure there are no logs for this batch
    shutil.rmtree(
        f"{opentaskpy.otflogging.LOG_DIRECTORY}/{task_id}", ignore_errors=True
    )

    config_loader = ConfigLoader("test/cfg")
    batch_obj = batch.Batch(
        None,
        task_id,
        parallel_batch_with_single_failure_definition,
        config_loader,
    )
    # Run and expect a false status
    assert not batch_obj.run()

    # Validate that a log has been created with the correct status
    # Use the logging module to get the right log file name
    log_file_name_batch = opentaskpy.otflogging._define_log_file_name(task_id, "B")
    log_file_name_touch_task = opentaskpy.otflogging._define_log_file_name("touch", "E")
    log_file_name_failed_task = opentaskpy.otflogging._define_log_file_name(
        "fail-command", "E"
    )

    # Check that all exist, with the right status
    assert os.path.exists(log_file_name_batch.replace("_running", "_failed"))
    assert os.path.exists(log_file_name_touch_task.replace("_running", ""))
    assert os.path.exists(log_file_name_failed_task.replace("_running", "_failed"))

    # Reset the prefix so it doesn't get reused
    del os.environ["OTF_LOG_RUN_PREFIX"]

    # Run it again, but this time, expect it to only run the failed task
    batch_obj = batch.Batch(
        None,
        task_id,
        parallel_batch_with_single_failure_definition,
        config_loader,
    )

    # Check the task_order_tree to only have the failed task in a NOT_STARTED state
    assert batch_obj.task_order_tree[1]["status"] == "COMPLETED"
    assert batch_obj.task_order_tree[2]["status"] == "NOT_STARTED"

    # Run and expect a false status
    assert not batch_obj.run()

    # Validate that the touch task has been skipped, so there's no log file
    log_file_name_batch = opentaskpy.otflogging._define_log_file_name(task_id, "B")
    log_file_name_touch_task = opentaskpy.otflogging._define_log_file_name("touch", "E")
    log_file_name_failed_task = opentaskpy.otflogging._define_log_file_name(
        "fail-command", "E"
    )

    # Check that all exist, with the right status
    assert os.path.exists(log_file_name_batch.replace("_running", "_failed"))
    assert not os.path.exists(log_file_name_touch_task)
    assert os.path.exists(log_file_name_failed_task.replace("_running", "_failed"))


def test_batch_resume_after_failure_retry_successful_tasks(
    setup_ssh_keys, env_vars, root_dir, clear_logs
):
    task_id = f"parallel-single-failure-2-{RANDOM}"
    # Ensure there are no logs for this batch
    shutil.rmtree(
        f"{opentaskpy.otflogging.LOG_DIRECTORY}/{task_id}", ignore_errors=True
    )

    config_loader = ConfigLoader("test/cfg")
    batch_obj = batch.Batch(
        None,
        task_id,
        parallel_batch_with_single_failure_with_retry_definition,
        config_loader,
    )
    # Run and expect a false status
    assert not batch_obj.run()

    # Validate that a log has been created with the correct status
    # Use the logging module to get the right log file name
    log_file_name_batch = opentaskpy.otflogging._define_log_file_name(task_id, "B")
    log_file_name_touch_task = opentaskpy.otflogging._define_log_file_name("touch", "E")
    log_file_name_failed_task = opentaskpy.otflogging._define_log_file_name(
        "fail-command", "E"
    )

    # Check that all exist, with the right status
    assert os.path.exists(log_file_name_batch.replace("_running", "_failed"))
    assert os.path.exists(log_file_name_touch_task.replace("_running", ""))
    assert os.path.exists(log_file_name_failed_task.replace("_running", "_failed"))

    # Reset the prefix so it doesn't get reused
    del os.environ["OTF_LOG_RUN_PREFIX"]

    # Run it again, but this time, expect it to only run the failed task
    batch_obj = batch.Batch(
        None,
        task_id,
        parallel_batch_with_single_failure_with_retry_definition,
        config_loader,
    )

    # Check the task_order_tree to check both tasks have a NOT_STARTED state
    assert batch_obj.task_order_tree[1]["status"], "NOT_STARTED"
    assert batch_obj.task_order_tree[2]["status"], "NOT_STARTED"

    # Run and expect a false status
    assert not batch_obj.run()

    # Validate that the touch task has been skipped, so there's no log file
    log_file_name_batch = opentaskpy.otflogging._define_log_file_name(task_id, "B")
    log_file_name_touch_task = opentaskpy.otflogging._define_log_file_name("touch", "E")
    log_file_name_failed_task = opentaskpy.otflogging._define_log_file_name(
        "fail-command", "E"
    )

    # Check that all exist, with the right status
    assert os.path.exists(log_file_name_batch.replace("_running", "_failed"))
    assert os.path.exists(log_file_name_touch_task.replace("_running", ""))
    assert os.path.exists(log_file_name_failed_task.replace("_running", "_failed"))


def test_batch_parallel_many(setup_ssh_keys, env_vars, root_dir, clear_logs):
    # Forcing a prefix makes it easy to identify log files, as well as
    # ensuring that any rerun logic doesn't get hit
    os.environ["OTF_LOG_RUN_PREFIX"] = f"testbatch_many_parallel_{RANDOM}"

    config_loader = ConfigLoader("test/cfg")
    batch_obj = batch.Batch(
        None, f"parallel-many-{RANDOM}", parallel_batch_many_definition, config_loader
    )
    # Run and expect a true status
    assert batch_obj.run()


def test_batch_continue_on_failure(setup_ssh_keys, env_vars, root_dir, clear_logs):
    task_id = f"dependency-continue-on-fail-1-{RANDOM}"
    # Ensure there are no logs for this batch
    shutil.rmtree(
        f"{opentaskpy.otflogging.LOG_DIRECTORY}/{task_id}", ignore_errors=True
    )

    config_loader = ConfigLoader("test/cfg")
    batch_obj = batch.Batch(
        None,
        task_id,
        dependent_batch_continue_on_fail_definition,
        config_loader,
    )
    # Run and expect a false status
    assert not batch_obj.run()

    # Validate that a log has been created with the correct status
    # Use the logging module to get the right log file name
    log_file_name_batch = opentaskpy.otflogging._define_log_file_name(task_id, "B")
    log_file_name_fail_task = opentaskpy.otflogging._define_log_file_name(
        "fail-command", "E"
    )
    log_file_name_scp_task = opentaskpy.otflogging._define_log_file_name(
        "scp-basic", "T"
    )

    # Check that all exist, with the right status
    assert os.path.exists(log_file_name_batch.replace("_running", "_failed"))

    # The failed task that we continue on, should still have a failed log
    assert os.path.exists(log_file_name_fail_task.replace("_running", "_failed"))

    # The successful task should have a successful
    assert os.path.exists(log_file_name_scp_task.replace("_running", ""))



# pylint: skip-file
import pytest

from opentaskpy.config.schemas import validate_transfer_json


@pytest.fixture(scope="function")
def valid_protocol_definition():
    return {
        "name": "ssh",
        "credentials": {
            "username": "test",
        },
    }


@pytest.fixture(scope="function")
def valid_source_definition(valid_protocol_definition):
    return {
        "hostname": "{{ HOST_A }}",
        "directory": "/tmp/testFiles/src",
        "fileRegex": ".*\\.txt",
        "protocol": valid_protocol_definition,
    }


@pytest.fixture(scope="function")
def valid_destination_definition(valid_protocol_definition):
    return {
        "hostname": "somehost",
        "directory": "/tmp/testFiles/dest",
        "protocol": valid_protocol_definition,
    }


def test_ssh_basic(valid_source_definition, valid_destination_definition):
    json_data = {
        "type": "transfer",
        "source": valid_source_definition,
        "destination": [],
    }
    # We dont actually need to give any destinations
    assert validate_transfer_json(json_data)

    # Add a valid destination
    json_data["destination"].append(valid_destination_definition)

    assert validate_transfer_json(json_data)

    # Remove protocol
    del json_data["destination"][0]["protocol"]
    assert not validate_transfer_json(json_data)


def test_ssh_rename(valid_source_definition, valid_destination_definition):
    json_data = {
        "type": "transfer",
        "source": valid_source_definition,
        "destination": [valid_destination_definition],
    }

    # Add rename config
    json_data["destination"][0]["rename"] = {"pattern": "test", "sub": "blah"}
    assert validate_transfer_json(json_data)


def test_ssh_permissions(valid_source_definition, valid_destination_definition):
    json_data = {
        "type": "transfer",
        "source": valid_source_definition,
        "destination": [valid_destination_definition],
    }

    # Add permission
    json_data["destination"][0]["permissions"] = {
        "owner": "test",
        "group": "blah",
        "mode": "0777",
    }
    assert validate_transfer_json(json_data)

    json_data["destination"][0]["permissions"] = {
        "mode": "0777",
    }

    assert validate_transfer_json(json_data)

    json_data["destination"][0]["permissions"] = {
        "invalid": "test",
    }
    assert not validate_transfer_json(json_data)


def test_ssh_flags(valid_source_definition, valid_protocol_definition):
    json_data = {
        "type": "transfer",
        "source": valid_source_definition,
        "destination": [
            {
                "hostname": "{{ HOST_A }}",
                "directory": "/tmp/testFiles/src",
                "flags": {},
                "protocol": valid_protocol_definition,
            }
        ],
    }
    assert not validate_transfer_json(json_data)

    # Add full path for flags
    json_data["destination"][0]["flags"]["fullPath"] = "/tmp/testFiles/src/flag.file"
    assert validate_transfer_json(json_data)


def test_ssh_transfer_type(valid_source_definition, valid_destination_definition):
    json_data = {
        "type": "transfer",
        "source": valid_source_definition,
        "destination": [valid_destination_definition],
    }

    # Add transfer type
    for type in ["push", "pull", "proxy"]:
        json_data["destination"][0]["transferType"] = type
        assert validate_transfer_json(json_data)

    json_data["destination"][0]["transferType"] = "blah"
    assert not validate_transfer_json(json_data)



"""Logging module."""

import json
import logging
import os
import re
from datetime import datetime

OTF_LOG_FORMAT = (
    "%(asctime)s — %(levelname)s - %(name)s - %(filename)s:%(lineno)s [%(threadName)s]"
    " — %(message)s"
)
LOG_DIRECTORY = (
    "logs"
    if os.environ.get("OTF_LOG_DIRECTORY") is None
    else os.environ.get("OTF_LOG_DIRECTORY")
)


class JSONFormatter(logging.Formatter):
    """JSON formatter.

    This will take a log message and output it as JSON rather than plain text

    Args:
        logging (logging.Formatter): The logging formatter to use
    """

    # Add an init that takes task_id
    def __init__(self, task_id: str | None):
        """Initialise the formatter.

        Args:
            task_id (str | None): The Task ID. Defaults to None.
        """
        super().__init__()
        self.task_id = task_id

    def format(self, record):
        """Format the log message as JSON."""
        task_id = self.task_id
        run_id = os.environ.get("OTF_RUN_ID")
        message = record.msg
        json_log_record = {
            "timestamp": datetime.utcnow().isoformat(),
            "level": record.levelname,
            "function": record.funcName,
            "file": f"{record.pathname}:{record.lineno}",
            "thread": record.threadName,
            "logger": record.name,
            "message": message,
            "task_id": task_id,
            "run_id": run_id,
        }
        record.msg = json.dumps(json_log_record)
        return super().format(record)


def _define_log_file_name(task_id: str | None, task_type: str | None) -> str:
    global LOG_DIRECTORY  # pylint: disable=global-statement
    LOG_DIRECTORY = (
        "logs"
        if os.environ.get("OTF_LOG_DIRECTORY") is None
        else os.environ.get("OTF_LOG_DIRECTORY")
    )

    # Set a custom handler to write to a specific file
    # Get the appropriate timestamp for the log file
    prefix = datetime.now().strftime("%Y%m%d-%H%M%S.%f")[:-3]
    if os.environ.get("OTF_LOG_RUN_PREFIX") is not None:
        # This might have already been set by task_run, so pull it
        # in from the environment
        prefix = f"{os.environ.get('OTF_LOG_RUN_PREFIX')}"
    else:
        os.environ["OTF_LOG_RUN_PREFIX"] = prefix

    if task_type:
        task_type = f"_{task_type}"
    else:
        task_type = ""

    directory = f"{LOG_DIRECTORY}"
    if os.environ.get("OTF_RUN_ID") is not None:
        directory = f"{directory}/{os.environ.get('OTF_RUN_ID')}"
        filename = f"{prefix}{task_type}_{task_id}_running.log"
    else:
        if task_id is None:
            task_id = "no_task_id"
        directory = f"{directory}/{task_id}"
        filename = f"{prefix}{task_type}_running.log"

    return f"{directory}/{filename}"


def init_logging(
    name: str, task_id: str | None = None, task_type: str | None = None
) -> logging.Logger:
    """Setup a logger with the custom format and output filename.

    Args:
        name (str): The name of the logger, usually the class name, but can also
        reference the thread, or batch task
        task_id (str | None, optional): The Task ID. Defaults to None.
        task_type (str | None, optional): The Task Type, either T for transfer, E for execution or B for batch. Defaults to None.

    Returns:
        logging.Logger: The logger object used for logging output.
    """
    # Set the log format
    formatter = logging.Formatter(OTF_LOG_FORMAT)

    # Check if there's a root logger already
    if not logging.getLogger().hasHandlers():
        sfh = logging.StreamHandler()

        formatter = (
            JSONFormatter(task_id)
            if os.environ.get("OTF_LOG_JSON") and os.environ.get("OTF_LOG_JSON") == "1"
            else formatter
        )

        sfh.setFormatter(formatter)

        # Set the root logger
        logging.basicConfig(
            format=OTF_LOG_FORMAT,
            level=logging.INFO,
            handlers=[
                sfh,
            ],
        )

    # Create a unique logger object for this task
    if not task_id:
        otf_logger = logging.getLogger(f"{name}")
    else:
        otf_logger = logging.getLogger(f"{name}.{task_id}")

    # Set verbosity
    otf_logger.setLevel(logging.getLogger().getEffectiveLevel())
    # Ensure the logger is at least at INFO level
    if otf_logger.getEffectiveLevel() > logging.INFO:
        otf_logger.setLevel(logging.INFO)

    # If the log level is set in the environment, then use that
    if os.environ.get("OTF_LOG_LEVEL") is not None:
        otf_logger.setLevel(os.environ["OTF_LOG_LEVEL"])

    # If OTF_NO_LOG is set, then don't create the handler
    if os.environ.get("OTF_NO_LOG") is not None or not task_id:
        return otf_logger

    log_file_name = _define_log_file_name(task_id, task_type)

    tfh = TaskFileHandler(log_file_name)
    # Check there are no handlers of this class already with the same baseFilename
    if not any(
        isinstance(handler, TaskFileHandler)
        and handler.baseFilename == tfh.baseFilename
        for handler in otf_logger.handlers
    ):
        otf_logger.addHandler(tfh)
        tfh.setFormatter(formatter)

    otf_logger.debug("Logging initialised")

    return otf_logger


def get_latest_log_file(task_id: str, task_type: str) -> str | None:
    """Get the latest log file for this task.

    Args:
        task_id (str): The Task ID
        task_type (str): The task type

    Returns:
        str | None: Either the latest log file name, or None if there are no previous
        logs
    """
    log_file_name = _define_log_file_name(task_id, task_type)
    # Obviously the date/time in the filename needs to be replaced with the latest
    # log file
    # Replace the prefix with a regex wildcard
    log_file_name = log_file_name.replace(os.environ["OTF_LOG_RUN_PREFIX"], ".*")
    # Also, we don't want to limit to running jobs, only failed or successful ones
    log_file_name = log_file_name.replace("_running", "(_failed)*")

    if not os.path.exists(os.path.dirname(log_file_name)):
        return None

    # List the contents of the directory
    log_files = os.listdir(os.path.dirname(log_file_name))
    # Filter the list to only include files that match the log_file_name, and contain a valid date/time prefix
    log_files = [
        f
        for f in log_files
        if re.match(os.path.basename(log_file_name), f)
        and re.match(r"\d{8}-\d{6}\.\d{3}", f.split("_")[0])
    ]

    # Unless another date is given, only look at today's logs
    batch_resume_date = datetime.now().date()
    if os.environ.get("OTF_BATCH_RESUME_LOG_DATE"):
        batch_resume_date = datetime.strptime(
            os.environ.get("OTF_BATCH_RESUME_LOG_DATE"), "%Y%m%d"
        ).date()

    # Remove any logs that are not from the given date
    log_files = [
        f
        for f in log_files
        if datetime.strptime(f.split("_")[0], "%Y%m%d-%H%M%S.%f").date()
        == batch_resume_date
    ]

    # Sort the list by the date/time in the filename
    log_files.sort(key=lambda x: datetime.strptime(x.split("_")[0], "%Y%m%d-%H%M%S.%f"))
    # Get the latest log file
    if log_files:
        log_file_name = f"{os.path.dirname(log_file_name)}/{log_files[-1]}"
        logger.info(f"Latest log file: {log_file_name}")
        # If the last log was a failure, return that, otherwise we just start from scratch, so return nothing
        if "_failed" in log_file_name:
            return log_file_name

        logger.info("No failed log file found. Starting from scratch.")

    return None


def _mkdir(path: str) -> None:
    os.makedirs(path, exist_ok=True)


def close_log_file(logger__: logging.Logger, result: bool = False) -> None:
    """Closes the log file handle.

    Closes the log file, and renames the log file using the handler, based on the result
    argument

    Args:
        logger__ (_type_): The logger that needs to be closed.
        result (bool, optional): The return status of the task that was run. Defaults to False.
    """
    log_file_name = None
    # Close the log file
    for handler in logger__.handlers:
        # If its a task file handler, and the log file still exists
        if isinstance(handler, TaskFileHandler) and os.path.exists(
            handler.baseFilename
        ):
            log_file_name = handler.baseFilename

    log_handlers = []

    if log_file_name:
        new_log_filename = None
        if result:
            new_log_filename = log_file_name.replace("_running", "")
        elif result is not None and not result:
            new_log_filename = log_file_name.replace("_running", "_failed")

        # Loop through every logger that exists and has a handler of this filename, and
        # call the close method on it. Only the last one should rename the file
        for logger_ in list(logging.Logger.manager.loggerDict.values()):
            if isinstance(logger_, logging.Logger):
                for handler in logger_.handlers:
                    if (
                        isinstance(handler, TaskFileHandler)
                        and handler.baseFilename == log_file_name
                    ):
                        log_handlers.append(handler)
                        handler.close()

        # Now everything is closed, we can rename the log file
        # If result is True, then rename the file and remove _running from the name
        if new_log_filename:
            os.rename(log_file_name, new_log_filename)

            # Change the basename of the handler to match the new filename, in case it
            # wants to log anything else
            for handler in log_handlers:
                handler.baseFilename = new_log_filename


logger = init_logging(__name__)


# mypy: ignore-errors
class TaskFileHandler(logging.FileHandler):
    """Custom file handler to ensure logs get correct naming.

    This class handles the closing of log files. It will rename based on the argument
    passed into the close method.
    """

    def __init__(self, filename, mode="a", encoding=None, delay=True):
        """Create the log directory.

        Overrides the default __init_ to ensure that the path containing the log exists
        before attempting to write to it.

        """
        _mkdir(os.path.dirname(filename))
        logging.FileHandler.__init__(self, filename, mode, encoding, delay)



"""Basic HTTP JSON lookup plugin.

Does an HTTP GET request to the given URL and returns the value of an attribute JSON response.
This does not support any type of authentication, and just expects a simple
JSON response from which to extract a value using a provided JSONPath.

"""

import requests
from jsonpath_ng import parse

import opentaskpy.otflogging
from opentaskpy.exceptions import LookupPluginError

logger = opentaskpy.otflogging.init_logging(__name__)

PLUGIN_NAME = "http_json"


def run(**kwargs) -> str | None:  # type: ignore[no-untyped-def]
    """Pull a variable from a URL that returns JSON.

    The endpoint must respond within 5 seconds, or the request will fail.

    Args:
        **kwargs: Expect a kwarg named url and jsonpath. This should be the URL to the
        JSON endpoint, and the JSONPath to the attribute to return.

    Raises:
        LookupPluginError: Returned if the kwarg 'url' or 'jsonpath' is not provided
        FileNotFoundError: Returned if the file does not exist

    Returns:
        _type_: The value read from the file
    """
    # Expect a kwarg named url, and value
    expected_kwargs = ["url", "jsonpath"]
    for kwarg in expected_kwargs:
        if kwarg not in kwargs:
            raise LookupPluginError(
                f"Missing kwarg: '{kwarg}' while trying to run lookup plugin"
                f" '{PLUGIN_NAME}'"
            )

    result: str | None = None
    try:
        response = requests.get(kwargs["url"], timeout=5)
        response.raise_for_status()
        json = response.json()
        jsonpath_expr = parse(kwargs["jsonpath"])
        result = [match.value for match in jsonpath_expr.find(json)][0]

        logger.log(12, f"Read '{result}' from URL {kwargs['url']}")
    except Exception as ex_:  # pylint: disable=broad-except
        logger.error(f"Failed to read from URL {kwargs['url']}: {ex_}")

    return result



"""Email handler to send files via email."""

import glob
import smtplib
from email.mime.application import MIMEApplication
from email.mime.multipart import MIMEMultipart
from email.mime.text import MIMEText

import opentaskpy.otflogging
from opentaskpy.remotehandlers.remotehandler import RemoteTransferHandler

MAX_OBJECTS_PER_QUERY = 100


class EmailTransfer(RemoteTransferHandler):
    """A remote handler for email transfers."""

    TASK_TYPE = "T"

    def __init__(self, spec: dict):
        """Initialise the handler."""
        self.protocol_vars: dict

        self.logger = opentaskpy.otflogging.init_logging(
            __name__, spec["task_id"], self.TASK_TYPE
        )

        super().__init__(spec)

    def supports_direct_transfer(self) -> bool:
        """Return whether this handler supports direct transfers."""
        return False

    def set_handler_vars(self, protocol_vars: dict) -> None:
        """Set the handler variables.

        Set any custom variables that are specifically needed for this handler.

        The schema defines the following variables:
        - smtp_server - The SMTP server to use
        - smtp_port - The SMTP port to use
        - sender - The sender of the email
        - credentials - A dictionary containing the following:
            - username - The username used to authenticate
            - password - The password used to authenticate

        Args:
            protocol_vars (dict): The protocol variables.

        """
        self.protocol_vars = protocol_vars

        # Also pull variables that have been set on the spec level
        # Replace/Append anything defined in self.protocol_vars with anything in self.spec["protocol"]
        for key, value in self.spec["protocol"].items():
            self.protocol_vars[key] = value
        # Remove name
        del self.protocol_vars["name"]

    def push_files_from_worker(
        self, local_staging_directory: str, file_list: dict | None = None
    ) -> int:
        """Push files from the worker to the email recipients.

        Args:
            local_staging_directory (str): The local staging directory.
            file_list (dict, optional): A dictionary of files to transfer. Defaults to None.

        Returns:
            int: The result of the transfer.
        """
        result = 0
        if file_list:
            files = list(file_list.keys())
        else:
            files = glob.glob(f"{local_staging_directory}/*")

        for email_address in self.spec["recipients"]:
            # Create an email message
            msg = MIMEMultipart()

            # Attach the files to the message
            for file in files:
                # Strip the directory from the file
                file_name = file.split("/")[-1]
                self.logger.debug(f"Emailing file: {files} to {email_address}")
                try:
                    with open(file, "rb") as file_handle:
                        part = MIMEApplication(file_handle.read(), Name=file_name)
                    # After the file is closed
                    part["Content-Disposition"] = f'attachment; filename="{file_name}"'
                    msg.attach(part)
                except Exception as ex:  # pylint: disable=broad-exception-caught
                    self.logger.error(f"Failed to attach file: {file}")
                    self.logger.error(ex)
                    result = 1

            # Get comma separated list of files
            attachment_file_list = ", ".join([file.split("/")[-1] for file in files])

            # Add a plaintext body to the email
            msg.attach(
                MIMEText(
                    self.spec.get(
                        "message", f"Please find attached: {attachment_file_list}"
                    )
                )
            )
            # Set the email subject
            if "subject" in self.spec:
                msg["Subject"] = self.spec["subject"]

            msg["From"] = self.protocol_vars["sender"]

            # Send the email using a provided SMTP server
            try:
                self.logger.debug(f"Sending email to {email_address}")
                smtp = smtplib.SMTP(
                    self.protocol_vars["smtp_server"],
                    port=self.protocol_vars["smtp_port"],
                )
                smtp.starttls()

                # Authenticate
                smtp.login(
                    self.protocol_vars["credentials"]["username"],
                    self.protocol_vars["credentials"]["password"],
                )

                smtp.sendmail(
                    self.protocol_vars["sender"], email_address, msg.as_string()
                )
                smtp.quit()
            except Exception as ex:  # pylint: disable=broad-exception-caught
                self.logger.error(f"Failed to send email to {email_address}")
                self.logger.error(ex)
                result = 1

        return result

    def pull_files_to_worker(self, local_staging_directory: str) -> int:
        """Not implemented for this handler."""
        raise NotImplementedError

    def handle_post_copy_action(self, files: list) -> None:
        """Not implemented for this handler."""
        raise NotImplementedError

    def list_files(
        self, directory: str | None = None, file_pattern: str | None = None
    ) -> dict:
        """Not implemented for this handler."""
        raise NotImplementedError

    def move_files_to_final_location(self, files: list) -> None:
        """Not implemented for this handler."""
        raise NotImplementedError

    def pull_files(self, files: list) -> None:
        """Not implemented for this handler."""
        raise NotImplementedError

    def transfer_files(
        self, files: dict, remote_spec: dict, dest_remote_handler: dict | None = None
    ) -> None:
        """Not implemented for this handler."""
        raise NotImplementedError
